diff --git a/core/lib/Drupal/Core/Datetime/Entity/DateFormat.php b/core/lib/Drupal/Core/Datetime/Entity/DateFormat.php
index bb4efe0aa8c3b730ac8982111c2c01437c83be53..5eb0faf5bf8e805bda256aa28d8f3624c8d9dbc5 100644
--- a/core/lib/Drupal/Core/Datetime/Entity/DateFormat.php
+++ b/core/lib/Drupal/Core/Datetime/Entity/DateFormat.php
@@ -25,6 +25,7 @@
  *     "label" = "label"
  *   },
  *   admin_permission = "administer site configuration",
+ *   list_cache_tags = { "rendered" }
  * )
  */
 class DateFormat extends ConfigEntityBase implements DateFormatInterface {
@@ -98,11 +99,4 @@ public function getCacheTag() {
     return ['rendered'];
   }
 
-  /**
-   * {@inheritdoc}
-   */
-  public function getListCacheTags() {
-    return ['rendered'];
-  }
-
 }
diff --git a/core/lib/Drupal/Core/Entity/Entity.php b/core/lib/Drupal/Core/Entity/Entity.php
index 07e50c7b6021266e2dc576312d258e996cfa935c..558297b66c9998dc83f17fdfeedd3b2c68e95eae 100644
--- a/core/lib/Drupal/Core/Entity/Entity.php
+++ b/core/lib/Drupal/Core/Entity/Entity.php
@@ -380,7 +380,6 @@ public function preSave(EntityStorageInterface $storage) {
    * {@inheritdoc}
    */
   public function postSave(EntityStorageInterface $storage, $update = TRUE) {
-    $this->onSaveOrDelete();
     $this->invalidateTagsOnSave($update);
   }
 
@@ -406,7 +405,7 @@ public static function preDelete(EntityStorageInterface $storage, array $entitie
    * {@inheritdoc}
    */
   public static function postDelete(EntityStorageInterface $storage, array $entities) {
-    self::invalidateTagsOnDelete($entities);
+    self::invalidateTagsOnDelete($storage->getEntityType(), $entities);
   }
 
   /**
@@ -426,15 +425,8 @@ public function referencedEntities() {
    * {@inheritdoc}
    */
   public function getCacheTag() {
-    return [$this->entityTypeId . ':' . $this->id()];
-  }
-
-  /**
-   * {@inheritdoc}
-   */
-  public function getListCacheTags() {
     // @todo Add bundle-specific listing cache tag? https://drupal.org/node/2145751
-    return [$this->entityTypeId . 's'];
+    return [$this->entityTypeId . ':' . $this->id()];
   }
 
   /**
@@ -461,26 +453,6 @@ public static function create(array $values = array()) {
     return $entity_manager->getStorage($entity_manager->getEntityTypeFromClass(get_called_class()))->create($values);
   }
 
-
-  /**
-   * Acts on an entity after it was saved or deleted.
-   */
-  protected function onSaveOrDelete() {
-    $referenced_entities = array(
-      $this->getEntityTypeId() => array($this->id() => $this),
-    );
-
-    foreach ($this->referencedEntities() as $referenced_entity) {
-      $referenced_entities[$referenced_entity->getEntityTypeId()][$referenced_entity->id()] = $referenced_entity;
-    }
-
-    foreach ($referenced_entities as $entity_type => $entities) {
-      if ($this->entityManager()->hasHandler($entity_type, 'view_builder')) {
-        $this->entityManager()->getViewBuilder($entity_type)->resetCache($entities);
-      }
-    }
-  }
-
   /**
    * Invalidates an entity's cache tags upon save.
    *
@@ -492,7 +464,7 @@ protected function invalidateTagsOnSave($update) {
     // updated entity may start to appear in a listing because it now meets that
     // listing's filtering requirements. A newly created entity may start to
     // appear in listings because it did not exist before.)
-    $tags = $this->getListCacheTags();
+    $tags = $this->getEntityType()->getListCacheTags();
     if ($update) {
       // An existing entity was updated, also invalidate its unique cache tag.
       $tags = Cache::mergeTags($tags, $this->getCacheTag());
@@ -504,19 +476,20 @@ protected function invalidateTagsOnSave($update) {
   /**
    * Invalidates an entity's cache tags upon delete.
    *
+   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
+   *   The entity type definition.
    * @param \Drupal\Core\Entity\EntityInterface[] $entities
    *   An array of entities.
    */
-  protected static function invalidateTagsOnDelete(array $entities) {
-    $tags = array();
+  protected static function invalidateTagsOnDelete(EntityTypeInterface $entity_type, array $entities) {
+    $tags = $entity_type->getListCacheTags();
     foreach ($entities as $entity) {
       // An entity was deleted: invalidate its own cache tag, but also its list
       // cache tags. (A deleted entity may cause changes in a paged list on
       // other pages than the one it's on. The one it's on is handled by its own
       // cache tag, but subsequent list pages would not be invalidated, hence we
       // must invalidate its list cache tags as well.)
-      $tags = Cache::mergeTags($tags, $entity->getCacheTag(), $entity->getListCacheTags());
-      $entity->onSaveOrDelete();
+      $tags = Cache::mergeTags($tags, $entity->getCacheTag());
     }
     Cache::invalidateTags($tags);
   }
diff --git a/core/lib/Drupal/Core/Entity/EntityInterface.php b/core/lib/Drupal/Core/Entity/EntityInterface.php
index 9280d79a4ce82450aa1a5395247f17deda0162fc..4023e10d0e95a7ebc5e51307eed166f8666a4fce 100644
--- a/core/lib/Drupal/Core/Entity/EntityInterface.php
+++ b/core/lib/Drupal/Core/Entity/EntityInterface.php
@@ -399,15 +399,4 @@ public function getTypedData();
    */
   public function getCacheTag();
 
-  /**
-   * The list cache tags associated with this entity.
-   *
-   * Enables code listing entities of this type to ensure that newly created
-   * entities show up immediately.
-   *
-   * @return array
-   *   An array of cache tags.
-   */
-  public function getListCacheTags();
-
 }
diff --git a/core/lib/Drupal/Core/Entity/EntityStorageInterface.php b/core/lib/Drupal/Core/Entity/EntityStorageInterface.php
index 7a329cc9af67b0bfc75f9461e739dd2e37e11759..082e9893b9774521f987bd0a4cfd7f4394f8bc32 100644
--- a/core/lib/Drupal/Core/Entity/EntityStorageInterface.php
+++ b/core/lib/Drupal/Core/Entity/EntityStorageInterface.php
@@ -45,7 +45,7 @@ public function resetCache(array $ids = NULL);
    * @param $ids
    *   An array of entity IDs, or NULL to load all entities.
    *
-   * @return array
+   * @return \Drupal\Core\Entity\EntityInterface[]
    *   An array of entity objects indexed by their IDs. Returns an empty array
    *   if no matching entities found.
    */
diff --git a/core/lib/Drupal/Core/Entity/EntityType.php b/core/lib/Drupal/Core/Entity/EntityType.php
index 8cc1cbe4b63c6a0f97fa81ec3d65cc4bcc4868ad..5ea863d94101523bce04a1f373ae08f8a4f827ac 100644
--- a/core/lib/Drupal/Core/Entity/EntityType.php
+++ b/core/lib/Drupal/Core/Entity/EntityType.php
@@ -202,6 +202,13 @@ class EntityType implements EntityTypeInterface {
    */
   protected $field_ui_base_route;
 
+  /**
+   * The list cache tags for this entity type.
+   *
+   * @var array
+   */
+  protected $list_cache_tags = array();
+
   /**
    * Constructs a new EntityType.
    *
@@ -234,6 +241,12 @@ public function __construct($definition) {
     $this->handlers += array(
       'access' => 'Drupal\Core\Entity\EntityAccessControlHandler',
     );
+
+    // Ensure a default list cache tag is set.
+    if (empty($this->list_cache_tags)) {
+      $this->list_cache_tags = [$definition['id'] . '_list'];
+    }
+
   }
 
   /**
@@ -660,4 +673,11 @@ public function getGroupLabel() {
     return !empty($this->group_label) ? (string) $this->group_label : $this->t('Other', array(), array('context' => 'Entity type group'));
   }
 
+  /**
+   * {@inheritdoc}
+   */
+  public function getListCacheTags() {
+    return $this->list_cache_tags;
+  }
+
 }
diff --git a/core/lib/Drupal/Core/Entity/EntityTypeInterface.php b/core/lib/Drupal/Core/Entity/EntityTypeInterface.php
index 40f5d5d31ab04d3ee4e8a4de9a7cc93b1295b313..f8237a5a37d6b26809d927259f6d33fc3d03eb4c 100644
--- a/core/lib/Drupal/Core/Entity/EntityTypeInterface.php
+++ b/core/lib/Drupal/Core/Entity/EntityTypeInterface.php
@@ -629,4 +629,13 @@ public function getUriCallback();
    */
   public function setUriCallback($callback);
 
+  /**
+   * The list cache tags associated with this entity type.
+   *
+   * Enables code listing entities of this type to ensure that newly created
+   * entities show up immediately.
+   *
+   * @return string[]
+   */
+  public function getListCacheTags();
 }
diff --git a/core/lib/Drupal/Core/Entity/EntityViewBuilder.php b/core/lib/Drupal/Core/Entity/EntityViewBuilder.php
index e24735a005a35ee9b4c9b5f43547e88f66310bd9..42f962b0ba88b9b444e7dfefc2f68ba655fe4cd1 100644
--- a/core/lib/Drupal/Core/Entity/EntityViewBuilder.php
+++ b/core/lib/Drupal/Core/Entity/EntityViewBuilder.php
@@ -349,11 +349,20 @@ public function getCacheTag() {
    * {@inheritdoc}
    */
   public function resetCache(array $entities = NULL) {
+    // If no set of specific entities is provided, invalidate the entity view
+    // builder's cache tag. This will invalidate all entities rendered by this
+    // view builder.
+    // Otherwise, if a set of specific entities is provided, invalidate those
+    // specific entities only, plus their list cache tags, because any lists in
+    // which these entities are rendered, must be invalidated as well. However,
+    // even in this case, we might invalidate more cache items than necessary.
+    // When we have a way to invalidate only those cache items that have both
+    // the individual entity's cache tag and the view builder's cache tag, we'll
+    // be able to optimize this further.
     if (isset($entities)) {
-      // Always invalidate the ENTITY_TYPE_list tag.
-      $tags = array($this->entityTypeId . '_list');
+      $tags = [];
       foreach ($entities as $entity) {
-        $tags = Cache::mergeTags($tags, $entity->getCacheTag());
+        $tags = Cache::mergeTags($tags, $entity->getCacheTag(), $entity->getEntityType()->getListCacheTags());
       }
       Cache::invalidateTags($tags);
     }
diff --git a/core/modules/aggregator/src/Entity/Item.php b/core/modules/aggregator/src/Entity/Item.php
index 0cebc3b32e76b692e611ffa9d36891e43be68ea8..09eed9b66d7d2c0776b5fa14167db5afd92395f2 100644
--- a/core/modules/aggregator/src/Entity/Item.php
+++ b/core/modules/aggregator/src/Entity/Item.php
@@ -31,6 +31,7 @@
  *   uri_callback = "Drupal\aggregator\Entity\Item::buildUri",
  *   base_table = "aggregator_item",
  *   render_cache = FALSE,
+ *   list_cache_tags = { "aggregator_feed_list" },
  *   entity_keys = {
  *     "id" = "iid",
  *     "label" = "title",
@@ -232,13 +233,6 @@ public function getCacheTag() {
     return Feed::load($this->getFeedId())->getCacheTag();
   }
 
-  /**
-   * {@inheritdoc}
-   */
-  public function getListCacheTags() {
-    return Feed::load($this->getFeedId())->getListCacheTags();
-  }
-
 
   /**
    * Entity URI callback.
diff --git a/core/modules/block/src/BlockViewBuilder.php b/core/modules/block/src/BlockViewBuilder.php
index d29b63c4efb6f8b6e71306f595a751f365e362de..32d75cff3425345991c05e5acf6546c107be7fb7 100644
--- a/core/modules/block/src/BlockViewBuilder.php
+++ b/core/modules/block/src/BlockViewBuilder.php
@@ -73,7 +73,6 @@ public function viewMultiple(array $entities = array(), $view_mode = 'full', $la
       $build[$entity_id]['#cache']['tags'] = Cache::mergeTags(
         $this->getCacheTag(), // Block view builder cache tag.
         $entity->getCacheTag(), // Block entity cache tag.
-        $entity->getListCacheTags(), // Block entity list cache tags.
         $plugin->getCacheTags() // Block plugin cache tags.
       );
 
diff --git a/core/modules/block/src/Entity/Block.php b/core/modules/block/src/Entity/Block.php
index 0036eb835fc79b2e33419e399d03bd9d85bc8b86..20d70a40cd3dcb6a5b38b3eca84d19cde33dc136 100644
--- a/core/modules/block/src/Entity/Block.php
+++ b/core/modules/block/src/Entity/Block.php
@@ -154,16 +154,33 @@ public function calculateDependencies() {
     return $this->dependencies;
   }
 
+  /**
+   * {@inheritdoc}
+   */
+  public function postSave(EntityStorageInterface $storage, $update = TRUE) {
+    parent::postSave($storage, $update);
+
+    // Entity::postSave() calls Entity::invalidateTagsOnSave(), which only
+    // handles the regular cases. The Block entity has one special case: a
+    // newly created block may *also* appear on any page in the current theme,
+    // so we must invalidate the associated block's cache tag (which includes
+    // the theme cache tag).
+    if (!$update) {
+      Cache::invalidateTags($this->getCacheTag());
+    }
+  }
+
   /**
    * {@inheritdoc}
    *
    * Block configuration entities are a special case: one block entity stores
-   * the placement of one block in one theme. Instead of using an entity type-
-   * specific list cache tag like most entities, use the cache tag of the theme
-   * this block is placed in instead.
+   * the placement of one block in one theme. Changing these entities may affect
+   * any page that is rendered in a certain theme, even if the block doesn't
+   * appear there currently. Hence a block configuration entity must also return
+   * the associated theme's cache tag.
    */
-  public function getListCacheTags() {
-    return array('theme:' . $this->theme);
+  public function getCacheTag() {
+    return Cache::mergeTags(parent::getCacheTag(), ['theme:' . $this->theme]);
   }
 
   /**
diff --git a/core/modules/book/book.module b/core/modules/book/book.module
index dc20b1d760a6949ec3e52560b4ac85480a25330c..6b631c7520172248062592f6279a95558d5bee23 100644
--- a/core/modules/book/book.module
+++ b/core/modules/book/book.module
@@ -14,6 +14,7 @@
 use Drupal\Core\Routing\RouteMatchInterface;
 use Drupal\node\NodeInterface;
 use Drupal\node\NodeTypeInterface;
+use Drupal\node\Entity\Node;
 use Drupal\Core\Language\LanguageInterface;
 use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
 use Drupal\Core\Template\Attribute;
@@ -243,6 +244,11 @@ function book_node_view(array &$build, EntityInterface $node, EntityViewDisplayI
             drupal_get_path('module', 'book') . '/css/book.theme.css',
           ),
         ),
+        // The book navigation is a listing of Node entities, so associate its
+        // list cache tag for correct invalidation.
+        '#cache' => [
+          'tags' => $node->getEntityType()->getListCacheTags(),
+        ],
       );
     }
   }
diff --git a/core/modules/book/src/BookExport.php b/core/modules/book/src/BookExport.php
index 4b4507804987214971a4208ce7c8ed0c5e04a738..b71c7a0874d1802011bcb0a23540e30a3ea01a54 100644
--- a/core/modules/book/src/BookExport.php
+++ b/core/modules/book/src/BookExport.php
@@ -8,6 +8,7 @@
 namespace Drupal\book;
 
 use Drupal\Core\Entity\EntityManagerInterface;
+use Drupal\node\Entity\Node;
 use Drupal\node\NodeInterface;
 
 /**
@@ -84,6 +85,9 @@ public function bookExportHtml(NodeInterface $node) {
       '#title' => $node->label(),
       '#contents' => $contents,
       '#depth' => $node->book['depth'],
+      '#cache' => [
+        'tags' => $node->getEntityType()->getListCacheTags(),
+      ],
     );
   }
 
diff --git a/core/modules/book/src/Controller/BookController.php b/core/modules/book/src/Controller/BookController.php
index 7a271b335cc50a1ee15d055cc52a71ee2af64f4b..9e7a11f4acc8fe0639a8ded3def423ddc96ec5b4 100644
--- a/core/modules/book/src/Controller/BookController.php
+++ b/core/modules/book/src/Controller/BookController.php
@@ -10,6 +10,7 @@
 use Drupal\book\BookExport;
 use Drupal\book\BookManagerInterface;
 use Drupal\Core\Controller\ControllerBase;
+use Drupal\node\Entity\Node;
 use Drupal\node\NodeInterface;
 use Symfony\Component\DependencyInjection\Container;
 use Symfony\Component\DependencyInjection\ContainerInterface;
@@ -115,6 +116,9 @@ public function bookRender() {
     return array(
       '#theme' => 'item_list',
       '#items' => $book_list,
+      '#cache' => [
+        'tags' => \Drupal::entityManager()->getDefinition('node')->getListCacheTags(),
+      ],
     );
   }
 
diff --git a/core/modules/comment/src/Plugin/Field/FieldFormatter/CommentDefaultFormatter.php b/core/modules/comment/src/Plugin/Field/FieldFormatter/CommentDefaultFormatter.php
index aa41d1223a6fd7c202f0ba45c5b97d7dff5d499c..02c310775bc382a23753b43f742dd4533aad6074 100644
--- a/core/modules/comment/src/Plugin/Field/FieldFormatter/CommentDefaultFormatter.php
+++ b/core/modules/comment/src/Plugin/Field/FieldFormatter/CommentDefaultFormatter.php
@@ -9,7 +9,9 @@
 
 use Drupal\comment\CommentManagerInterface;
 use Drupal\comment\CommentStorageInterface;
+use Drupal\comment\Entity\Comment;
 use Drupal\comment\Plugin\Field\FieldType\CommentItemInterface;
+use Drupal\Core\Entity\EntityManagerInterface;
 use Drupal\Core\Entity\EntityViewBuilderInterface;
 use Drupal\Core\Entity\EntityFormBuilderInterface;
 use Drupal\Core\Field\FieldItemListInterface;
@@ -67,6 +69,13 @@ public static function defaultSettings() {
    */
   protected $viewBuilder;
 
+  /**
+   * The entity manager.
+   *
+   * @var \Drupal\Core\Entity\EntityManagerInterface
+   */
+  protected $entityManager;
+
   /**
    * The entity form builder.
    *
@@ -87,8 +96,7 @@ public static function create(ContainerInterface $container, array $configuratio
       $configuration['view_mode'],
       $configuration['third_party_settings'],
       $container->get('current_user'),
-      $container->get('entity.manager')->getStorage('comment'),
-      $container->get('entity.manager')->getViewBuilder('comment'),
+      $container->get('entity.manager'),
       $container->get('entity.form_builder')
     );
   }
@@ -112,18 +120,17 @@ public static function create(ContainerInterface $container, array $configuratio
    *   Third party settings.
    * @param \Drupal\Core\Session\AccountInterface $current_user
    *   The current user.
-   * @param \Drupal\comment\CommentStorageInterface $comment_storage
-   *   The comment storage.
-   * @param \Drupal\Core\Entity\EntityViewBuilderInterface $comment_view_builder
-   *   The comment view builder.
+   * @param \Drupal\Core\Entity\EntityManagerInterface $entity_manager
+   *   The entity manager
    * @param \Drupal\Core\Entity\EntityFormBuilderInterface $entity_form_builder
    *   The entity form builder.
    */
-  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, AccountInterface $current_user, CommentStorageInterface $comment_storage, EntityViewBuilderInterface $comment_view_builder, EntityFormBuilderInterface $entity_form_builder) {
+  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, AccountInterface $current_user, EntityManagerInterface $entity_manager, EntityFormBuilderInterface $entity_form_builder) {
     parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
-    $this->viewBuilder = $comment_view_builder;
-    $this->storage = $comment_storage;
+    $this->viewBuilder = $entity_manager->getViewBuilder('comment');
+    $this->storage = $entity_manager->getStorage('comment');
     $this->currentUser = $current_user;
+    $this->entityManager = $entity_manager;
     $this->entityFormBuilder = $entity_form_builder;
   }
 
@@ -150,19 +157,24 @@ public function viewElements(FieldItemListInterface $items) {
       // Unpublished comments are not included in
       // $entity->get($field_name)->comment_count, but unpublished comments
       // should display if the user is an administrator.
-      if ((($entity->get($field_name)->comment_count && $this->currentUser->hasPermission('access comments')) ||
-        $this->currentUser->hasPermission('administer comments'))) {
-        $mode = $comment_settings['default_mode'];
-        $comments_per_page = $comment_settings['per_page'];
-        $comments = $this->storage->loadThread($entity, $field_name, $mode, $comments_per_page, $this->getSetting('pager_id'));
-        if ($comments) {
-          comment_prepare_thread($comments);
-          $build = $this->viewBuilder->viewMultiple($comments);
-          $build['pager']['#theme'] = 'pager';
-          if ($this->getSetting('pager_id')) {
-            $build['pager']['#element'] = $this->getSetting('pager_id');
+      if ($this->currentUser->hasPermission('access comments') || $this->currentUser->hasPermission('administer comments')) {
+        // This is a listing of Comment entities, so associate its list cache
+        // tag for correct invalidation.
+        $output['comments']['#cache']['tags'] = $this->entityManager->getDefinition('comment')->getListCacheTags();
+
+        if ($entity->get($field_name)->comment_count || $this->currentUser->hasPermission('administer comments')) {
+          $mode = $comment_settings['default_mode'];
+          $comments_per_page = $comment_settings['per_page'];
+          $comments = $this->storage->loadThread($entity, $field_name, $mode, $comments_per_page, $this->getSetting('pager_id'));
+          if ($comments) {
+            comment_prepare_thread($comments);
+            $build = $this->viewBuilder->viewMultiple($comments);
+            $build['pager']['#theme'] = 'pager';
+            if ($this->getSetting('pager_id')) {
+              $build['pager']['#element'] = $this->getSetting('pager_id');
+            }
+            $output['comments'] += $build;
           }
-          $output['comments'] = $build;
         }
       }
 
diff --git a/core/modules/comment/src/Tests/CommentDefaultFormatterCacheTagsTest.php b/core/modules/comment/src/Tests/CommentDefaultFormatterCacheTagsTest.php
index 9c1468541b8ef9b7ace964757eaac7eb72751953..8747737f2d872de0506a8d098ac10568fd2673e8 100644
--- a/core/modules/comment/src/Tests/CommentDefaultFormatterCacheTagsTest.php
+++ b/core/modules/comment/src/Tests/CommentDefaultFormatterCacheTagsTest.php
@@ -68,6 +68,7 @@ public function testCacheTags() {
     $expected_cache_tags = array(
       'entity_test_view',
       'entity_test:'  . $commented_entity->id(),
+      'comment_list',
     );
     sort($expected_cache_tags);
     $this->assertEqual($build['#cache']['tags'], $expected_cache_tags, 'The test entity has the expected cache tags before it has comments.');
@@ -96,7 +97,7 @@ public function testCacheTags() {
     //   https://drupal.org/node/597236 lands, it's a temporary work-around.
     $commented_entity = entity_load('entity_test', $commented_entity->id(), TRUE);
 
-    // Verify cache tags on the rendered entity before it has comments.
+    // Verify cache tags on the rendered entity when it has comments.
     $build = \Drupal::entityManager()
       ->getViewBuilder('entity_test')
       ->view($commented_entity);
@@ -104,6 +105,7 @@ public function testCacheTags() {
     $expected_cache_tags = array(
       'entity_test_view',
       'entity_test:' . $commented_entity->id(),
+      'comment_list',
       'comment_view',
       'comment:' . $comment->id(),
       'filter_format:plain_text',
diff --git a/core/modules/comment/tests/src/Unit/Entity/CommentLockTest.php b/core/modules/comment/tests/src/Unit/Entity/CommentLockTest.php
index aef189e7cf183d8f82d9a348b7726ebbe32b3c52..c6f11bcb1b02c1e158118cad70a19a133eab2814 100644
--- a/core/modules/comment/tests/src/Unit/Entity/CommentLockTest.php
+++ b/core/modules/comment/tests/src/Unit/Entity/CommentLockTest.php
@@ -49,8 +49,8 @@ public function testLocks() {
     $methods = get_class_methods('Drupal\comment\Entity\Comment');
     unset($methods[array_search('preSave', $methods)]);
     unset($methods[array_search('postSave', $methods)]);
-    $methods[] = 'onSaveOrDelete';
     $methods[] = 'onUpdateBundleEntity';
+    $methods[] = 'invalidateTagsOnSave';
     $comment = $this->getMockBuilder('Drupal\comment\Entity\Comment')
       ->disableOriginalConstructor()
       ->setMethods($methods)
@@ -79,12 +79,6 @@ public function testLocks() {
       ->method('get')
       ->with('status')
       ->will($this->returnValue((object) array('value' => NULL)));
-    $comment->expects($this->once())
-      ->method('getCacheTag')
-      ->will($this->returnValue(array('comment:' . $cid)));
-    $comment->expects($this->once())
-      ->method('getListCacheTags')
-      ->will($this->returnValue(array('comments')));
     $storage = $this->getMock('Drupal\comment\CommentStorageInterface');
 
     // preSave() should acquire the lock. (This is what's really being tested.)
diff --git a/core/modules/filter/filter.module b/core/modules/filter/filter.module
index 9040b56f81a685e222a12c0337dea7c1c0c7c678..cad200c202a695ea01eb3f3ec65426aff4971e33 100644
--- a/core/modules/filter/filter.module
+++ b/core/modules/filter/filter.module
@@ -14,6 +14,7 @@
 use Drupal\Core\Routing\RouteMatchInterface;
 use Drupal\Core\Session\AccountInterface;
 use Drupal\Core\Template\Attribute;
+use Drupal\filter\Entity\FilterFormat;
 use Drupal\filter\FilterFormatInterface;
 
 /**
@@ -139,7 +140,7 @@ function filter_formats(AccountInterface $account = NULL) {
     else {
       $formats['all'] = \Drupal::entityManager()->getStorage('filter_format')->loadByProperties(array('status' => TRUE));
       uasort($formats['all'], 'Drupal\Core\Config\Entity\ConfigEntityBase::sort');
-      \Drupal::cache()->set("filter_formats:{$language_interface->id}", $formats['all'], Cache::PERMANENT, array('filter_formats'));
+      \Drupal::cache()->set("filter_formats:{$language_interface->id}", $formats['all'], Cache::PERMANENT, \Drupal::entityManager()->getDefinition('filter_format')->getListCacheTags());
     }
   }
 
diff --git a/core/modules/filter/src/FilterPluginManager.php b/core/modules/filter/src/FilterPluginManager.php
index 43b6568644c33ed40c457b06d8a9b520026168ef..ccbaf87b0801d39965c98cd86c6254c9774def8f 100644
--- a/core/modules/filter/src/FilterPluginManager.php
+++ b/core/modules/filter/src/FilterPluginManager.php
@@ -37,7 +37,7 @@ class FilterPluginManager extends DefaultPluginManager implements FallbackPlugin
   public function __construct(\Traversable $namespaces, CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler) {
     parent::__construct('Plugin/Filter', $namespaces, $module_handler, 'Drupal\filter\Plugin\FilterInterface', 'Drupal\filter\Annotation\Filter');
     $this->alterInfo('filter_info');
-    $this->setCacheBackend($cache_backend, 'filter_plugins', array('filter_formats'));
+    $this->setCacheBackend($cache_backend, 'filter_plugins');
   }
 
   /**
diff --git a/core/modules/hal/src/Tests/EntityTest.php b/core/modules/hal/src/Tests/EntityTest.php
index 591f7ffa2734c25b8e0ca743ef6cdfde79cc9e7a..47aa4d3a0901c0d18caed4347ef4559e4e2f86c0 100644
--- a/core/modules/hal/src/Tests/EntityTest.php
+++ b/core/modules/hal/src/Tests/EntityTest.php
@@ -64,7 +64,8 @@ public function testNode() {
       'body' => array(
         'value' => $this->randomMachineName(),
         'format' => $this->randomMachineName(),
-      )
+      ),
+      'revision_log' => $this->randomString(),
     ));
     $node->save();
 
@@ -160,13 +161,32 @@ public function testComment() {
     ));
     $node->save();
 
+    $parent_comment = entity_create('comment', array(
+      'uid' => $user->id(),
+      'subject' => $this->randomMachineName(),
+      'comment_body' => [
+        'value' => $this->randomMachineName(),
+        'format' => NULL,
+      ],
+      'entity_id' => $node->id(),
+      'entity_type' => 'node',
+      'field_name' => 'comment',
+    ));
+    $parent_comment->save();
+
     $comment = entity_create('comment', array(
       'uid' => $user->id(),
       'subject' => $this->randomMachineName(),
-      'comment_body' => $this->randomMachineName(),
+      'comment_body' => [
+        'value' => $this->randomMachineName(),
+        'format' => NULL,
+      ],
       'entity_id' => $node->id(),
       'entity_type' => 'node',
-      'field_name' => 'comment'
+      'field_name' => 'comment',
+      'pid' => $parent_comment->id(),
+      'mail' => 'dries@drupal.org',
+      'homepage' => 'http://buytaert.net',
     ));
     $comment->save();
 
diff --git a/core/modules/shortcut/src/Entity/Shortcut.php b/core/modules/shortcut/src/Entity/Shortcut.php
index 4bd041efa8c3e438348f07915b0ab4876a6676dd..75d5b2283d51a45cb6e9ec1c9017a607c944e325 100644
--- a/core/modules/shortcut/src/Entity/Shortcut.php
+++ b/core/modules/shortcut/src/Entity/Shortcut.php
@@ -46,6 +46,7 @@
  *     "delete-form" = "entity.shortcut.delete_form",
  *     "edit-form" = "entity.shortcut.canonical",
  *   },
+ *   list_cache_tags = { "shortcut_set_list" },
  *   bundle_entity_type = "shortcut_set"
  * )
  */
@@ -235,11 +236,4 @@ public function getCacheTag() {
     return $this->shortcut_set->entity->getCacheTag();
   }
 
-  /**
-   * {@inheritdoc}
-   */
-  public function getListCacheTags() {
-    return $this->shortcut_set->entity->getListCacheTags();
-  }
-
 }
diff --git a/core/modules/system/core.api.php b/core/modules/system/core.api.php
index 1863c287caba31c091b0c0f5652bdff7efdc76d0..9dc75de61782dc2a5ae733bfa61c74265b74b043 100644
--- a/core/modules/system/core.api.php
+++ b/core/modules/system/core.api.php
@@ -484,7 +484,7 @@
  * exact same cache tag invalidation as any of the built-in entity types, with
  * the ability to override any of the default behavior if needed.
  * See \Drupal\Core\Entity\EntityInterface::getCacheTag(),
- * \Drupal\Core\Entity\EntityInterface::getListCacheTags(),
+ * \Drupal\Core\Entity\EntityTypeInterface::getListCacheTags(),
  * \Drupal\Core\Entity\Entity::invalidateTagsOnSave() and
  * \Drupal\Core\Entity\Entity::invalidateTagsOnDelete().
  *
diff --git a/core/modules/system/src/Tests/Cache/PageCacheTagsTestBase.php b/core/modules/system/src/Tests/Cache/PageCacheTagsTestBase.php
index e54f97b55dd7443c4d58d6c81b3a0ae5098b9197..8d95233c5d9bbc57fb49b957bcc5cfbd47248020 100644
--- a/core/modules/system/src/Tests/Cache/PageCacheTagsTestBase.php
+++ b/core/modules/system/src/Tests/Cache/PageCacheTagsTestBase.php
@@ -15,6 +15,13 @@
  */
 abstract class PageCacheTagsTestBase extends WebTestBase {
 
+  /**
+   * {@inheritdoc}
+   *
+   * Always enable header dumping in page cache tags tests, this aids debugging.
+   */
+  protected $dumpHeaders = TRUE;
+
   /**
    * {@inheritdoc}
    */
@@ -50,6 +57,7 @@ protected function verifyPageCache($path, $hit_or_miss, $tags = FALSE) {
       $cid = sha1(implode(':', $cid_parts));
       $cache_entry = \Drupal::cache('render')->get($cid);
       sort($cache_entry->tags);
+      $tags = array_unique($tags);
       sort($tags);
       $this->assertIdentical($cache_entry->tags, $tags);
     }
diff --git a/core/modules/system/src/Tests/Entity/EntityViewBuilderTest.php b/core/modules/system/src/Tests/Entity/EntityViewBuilderTest.php
index 547663ce3967080c0fecb77995050ad51b681094..d57b79c663cf0fdb73bdec92e70e5371e343e20b 100644
--- a/core/modules/system/src/Tests/Entity/EntityViewBuilderTest.php
+++ b/core/modules/system/src/Tests/Entity/EntityViewBuilderTest.php
@@ -87,6 +87,7 @@ public function testEntityViewBuilderCacheWithReferences() {
     // Create an entity reference field and an entity that will be referenced.
     entity_reference_create_field('entity_test', 'entity_test', 'reference_field', 'Reference', 'entity_test');
     entity_get_display('entity_test', 'entity_test', 'full')->setComponent('reference_field', [
+      'type' => 'entity_reference_entity_view',
       'settings' => ['link' => FALSE],
     ])->save();
     $entity_test_reference = $this->createTestEntity('entity_test');
@@ -108,6 +109,7 @@ public function testEntityViewBuilderCacheWithReferences() {
     // Create another entity that references the first one.
     $entity_test = $this->createTestEntity('entity_test');
     $entity_test->reference_field->entity = $entity_test_reference;
+    $entity_test->reference_field->access = TRUE;
     $entity_test->save();
 
     // Get a fully built entity view render array.
@@ -124,7 +126,7 @@ public function testEntityViewBuilderCacheWithReferences() {
     $this->assertTrue($this->container->get('cache.' . $bin)->get($cid), 'The entity render element has been cached.');
 
     // Save the entity and verify that both cache entries have been deleted.
-    $entity_test->save();
+    $entity_test_reference->save();
     $this->assertFalse($this->container->get('cache.' . $bin)->get($cid), 'The entity render cache has been cleared when the entity was deleted.');
     $this->assertFalse($this->container->get('cache.' . $bin_reference)->get($cid_reference), 'The entity render cache for the referenced entity has been cleared when the entity was deleted.');
 
diff --git a/core/modules/system/tests/modules/entity_test/entity_test.routing.yml b/core/modules/system/tests/modules/entity_test/entity_test.routing.yml
index 25c159d1bc23ea4603a3c0f1ba447657780aa1c9..e3d0adca2faa6bf908b8ed3227d31c3885bd1df5 100644
--- a/core/modules/system/tests/modules/entity_test/entity_test.routing.yml
+++ b/core/modules/system/tests/modules/entity_test/entity_test.routing.yml
@@ -60,5 +60,22 @@ entity.entity_test.list_referencing_entities:
   requirements:
     _access: 'TRUE'
 
+entity.entity_test.list_labels_alphabetically:
+  path: '/entity_test/list_labels_alphabetically/{entity_type_id}'
+  defaults:
+    _content: '\Drupal\entity_test\Controller\EntityTestController::listEntitiesAlphabetically'
+    _title: 'List labels of entities of the given entity type alphabetically'
+  requirements:
+    _access: 'TRUE'
+
+entity.entity_test.list_empty:
+  path: '/entity_test/list_empty/{entity_type_id}'
+  defaults:
+    _content: '\Drupal\entity_test\Controller\EntityTestController::listEntitiesEmpty'
+    _title: 'Empty list of entities of the given entity type, empty because no entities match the query'
+  requirements:
+    _access: 'TRUE'
+
+
 route_callbacks:
   - '\Drupal\entity_test\Routing\EntityTestRoutes::routes'
diff --git a/core/modules/system/tests/modules/entity_test/src/Controller/EntityTestController.php b/core/modules/system/tests/modules/entity_test/src/Controller/EntityTestController.php
index a9d114aaa6edfe23d3698066af07eb82bda5320f..d189747f4e54dcc2eae5cff4f515da5d0aa14a38 100644
--- a/core/modules/system/tests/modules/entity_test/src/Controller/EntityTestController.php
+++ b/core/modules/system/tests/modules/entity_test/src/Controller/EntityTestController.php
@@ -7,6 +7,7 @@
 
 namespace Drupal\entity_test\Controller;
 
+use Drupal\Core\Cache\Cache;
 use Drupal\Core\Controller\ControllerBase;
 use Drupal\Core\Entity\Query\QueryFactory;
 use Symfony\Component\DependencyInjection\ContainerInterface;
@@ -121,4 +122,71 @@ public function listReferencingEntities($entity_reference_field_name, $reference
       ->viewMultiple($entities, 'full');
   }
 
+  /**
+   * List entities of the given entity type labels, sorted alphabetically.
+   *
+   * @param string $entity_type_id
+   *   The type of the entity being listed.
+   *
+   * @return array
+   *   A renderable array.
+   */
+  public function listEntitiesAlphabetically($entity_type_id) {
+    $entity_type_definition = $this->entityManager()->getDefinition($entity_type_id);
+    $query = $this->entityQueryFactory->get($entity_type_id);
+
+    // Sort by label field, if any.
+    if ($label_field = $entity_type_definition->getKey('label')) {
+      $query->sort($label_field);
+    }
+
+    $entities = $this->entityManager()
+      ->getStorage($entity_type_id)
+      ->loadMultiple($query->execute());
+
+    $cache_tags = [];
+    $labels = [];
+    foreach ($entities as $entity) {
+      $labels[] = $entity->label();
+      $cache_tags = Cache::mergeTags($cache_tags, $entity->getCacheTag());
+    }
+    // Always associate the list cache tag, otherwise the cached empty result
+    // wouldn't be invalidated. This would continue to show nothing matches the
+    // query, even though a newly created entity might match the query.
+    $cache_tags = Cache::mergeTags($cache_tags, $entity_type_definition->getListCacheTags());
+
+    return [
+      '#theme' => 'item_list',
+      '#items' => $labels,
+      '#title' => $entity_type_id . ' entities',
+      '#cache' => [
+        'tags' => $cache_tags,
+      ],
+    ];
+  }
+
+
+  /**
+   * Empty list of entities of the given entity type.
+   *
+   * Empty because no entities match the query. That may seem contrived, but it
+   * is an excellent way for testing whether an entity's list cache tags are
+   * working as expected.
+   *
+   * @param string $entity_type_id
+   *   The type of the entity being listed.
+   *
+   * @return array
+   *   A renderable array.
+   */
+  public function listEntitiesEmpty($entity_type_id) {
+    return [
+      '#theme' => 'item_list',
+      '#items' => [],
+      '#cache' => [
+        'tags' => $this->entityManager()->getDefinition($entity_type_id)->getListCacheTags(),
+      ],
+    ];
+  }
+
 }
diff --git a/core/modules/toolbar/toolbar.module b/core/modules/toolbar/toolbar.module
index 815d1cf11928c9c3965bf771284b3744265fcc31..2fbf64dc6436a9ba258edfc5b88c463a18b99e82 100644
--- a/core/modules/toolbar/toolbar.module
+++ b/core/modules/toolbar/toolbar.module
@@ -13,8 +13,7 @@
 use Drupal\Component\Datetime\DateTimePlus;
 use Drupal\Component\Utility\Crypt;
 use Drupal\Component\Utility\String;
-use Drupal\user\RoleInterface;
-use Drupal\user\UserInterface;
+use Drupal\user\Entity\Role;
 
 /**
  * Implements hook_help().
@@ -371,7 +370,8 @@ function _toolbar_get_subtrees_hash($langcode) {
     // caches later, based on the user's ID regardless of language.
     // Clear the cache when the 'locale' tag is deleted. This ensures a fresh
     // subtrees rendering when string translations are made.
-    \Drupal::cache('toolbar')->set($cid, $hash, Cache::PERMANENT, array('user:' . $uid, 'locale', 'menu:admin', 'user_roles'));
+    $role_list_cache_tags = \Drupal::entityManager()->getDefinition('user_role')->getListCacheTags();
+    \Drupal::cache('toolbar')->set($cid, $hash, Cache::PERMANENT, Cache::mergeTags(array('user:' . $uid, 'locale', 'menu:admin'), $role_list_cache_tags));
   }
   return $hash;
 }
diff --git a/core/modules/user/src/Tests/UserCacheTagsTest.php b/core/modules/user/src/Tests/UserCacheTagsTest.php
index d3ac2948af8925756e15da0739f206967b0a6d6f..4085de4f3f52dd13a2f3a8b19ab84507afe93bc2 100644
--- a/core/modules/user/src/Tests/UserCacheTagsTest.php
+++ b/core/modules/user/src/Tests/UserCacheTagsTest.php
@@ -48,4 +48,11 @@ protected function createEntity() {
     return $user;
   }
 
+  /**
+   * {@inheritdoc}
+   */
+  protected function getAdditionalCacheTagsForEntityListing() {
+    return ['user:0', 'user:1'];
+  }
+
 }
diff --git a/core/modules/user/src/Tests/UserPictureTest.php b/core/modules/user/src/Tests/UserPictureTest.php
index d20ade7361ff895ab0e8ac3c85054c4d92c2ffcf..639c38269acec5e895eebfd839113319eaebdc46 100644
--- a/core/modules/user/src/Tests/UserPictureTest.php
+++ b/core/modules/user/src/Tests/UserPictureTest.php
@@ -7,6 +7,7 @@
 
 namespace Drupal\user\Tests;
 
+use Drupal\Core\Cache\Cache;
 use Drupal\simpletest\WebTestBase;
 
 /**
@@ -102,6 +103,9 @@ function testPictureOnNodeComment() {
       ->set('features.comment_user_picture', TRUE)
       ->save();
 
+    // @todo Remove when https://www.drupal.org/node/2040135 lands.
+    Cache::invalidateTags(['rendered']);
+
     $edit = array(
       'comment_body[0][value]' => $this->randomString(),
     );
@@ -113,7 +117,9 @@ function testPictureOnNodeComment() {
       ->set('features.node_user_picture', FALSE)
       ->set('features.comment_user_picture', FALSE)
       ->save();
-    \Drupal::entityManager()->getViewBuilder('comment')->resetCache();
+
+    // @todo Remove when https://www.drupal.org/node/2040135 lands.
+    Cache::invalidateTags(['rendered']);
 
     $this->drupalGet('node/' . $node->id());
     $this->assertNoRaw(file_uri_target($file->getFileUri()), 'User picture not found on node and comment.');
diff --git a/core/modules/views_ui/src/ViewUI.php b/core/modules/views_ui/src/ViewUI.php
index 86d535b7914ad5760d853f5e9c40bc4e591befaf..260a71e843805f1d639ea71bb176e2685bc1f44d 100644
--- a/core/modules/views_ui/src/ViewUI.php
+++ b/core/modules/views_ui/src/ViewUI.php
@@ -1127,13 +1127,6 @@ public function getCacheTag() {
     $this->storage->getCacheTag();
   }
 
-  /**
-   * {@inheritdoc}
-   */
-  public function getListCacheTags() {
-    $this->storage->getListCacheTags();
-  }
-
   /**
    * {@inheritdoc}
    */
diff --git a/core/tests/Drupal/Tests/Core/Config/Entity/ConfigEntityStorageTest.php b/core/tests/Drupal/Tests/Core/Config/Entity/ConfigEntityStorageTest.php
index 6bf678fd0033939700ac9f925dcd020d78b27a80..4193f42fada47182e25c5fe42aa8628c574afb38 100644
--- a/core/tests/Drupal/Tests/Core/Config/Entity/ConfigEntityStorageTest.php
+++ b/core/tests/Drupal/Tests/Core/Config/Entity/ConfigEntityStorageTest.php
@@ -121,7 +121,9 @@ protected function setUp() {
     $this->entityType->expects($this->any())
       ->method('getClass')
       ->will($this->returnValue(get_class($this->getMockEntity())));
-
+    $this->entityType->expects($this->any())
+      ->method('getListCacheTags')
+      ->willReturn(array('test_entity_type_list'));
 
     $this->moduleHandler = $this->getMock('Drupal\Core\Extension\ModuleHandlerInterface');
 
@@ -241,7 +243,7 @@ public function testSaveInsert(EntityInterface $entity) {
     $this->cacheBackend->expects($this->once())
       ->method('invalidateTags')
       ->with(array(
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
 
     $this->configFactory->expects($this->exactly(2))
@@ -301,7 +303,7 @@ public function testSaveUpdate(EntityInterface $entity) {
       ->method('invalidateTags')
       ->with(array(
         $this->entityTypeId . ':foo', // Own cache tag.
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
 
     $this->configFactory->expects($this->exactly(2))
@@ -361,7 +363,7 @@ public function testSaveRename(ConfigEntityInterface $entity) {
       ->method('invalidateTags')
       ->with(array(
         $this->entityTypeId .':bar', // Own cache tag.
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
 
     $this->configFactory->expects($this->once())
@@ -493,7 +495,7 @@ public function testSaveNoMismatch() {
     $this->cacheBackend->expects($this->once())
       ->method('invalidateTags')
       ->with(array(
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
 
     $this->configFactory->expects($this->once())
@@ -728,7 +730,7 @@ public function testDelete() {
       ->with(array(
         $this->entityTypeId . ':bar', // Own cache tag.
         $this->entityTypeId . ':foo', // Own cache tag.
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
 
     $this->configFactory->expects($this->exactly(2))
@@ -790,7 +792,6 @@ public function testDeleteNothing() {
    * @return \Drupal\Core\Entity\EntityInterface|\PHPUnit_Framework_MockObject_MockObject
    */
   public function getMockEntity(array $values = array(), $methods = array()) {
-    $methods[] = 'onSaveOrDelete';
     $methods[] = 'onUpdateBundleEntity';
     return $this->getMockForAbstractClass('Drupal\Core\Config\Entity\ConfigEntityBase', array($values, 'test_entity_type'), '', TRUE, TRUE, TRUE, $methods);
   }
diff --git a/core/tests/Drupal/Tests/Core/Entity/EntityUnitTest.php b/core/tests/Drupal/Tests/Core/Entity/EntityUnitTest.php
index 1fc8d7b1c9ab48930e35d2d9e4fd70a4ba8988d8..a2034e78cfa2a42b111213ad79f8d56ae960088a 100644
--- a/core/tests/Drupal/Tests/Core/Entity/EntityUnitTest.php
+++ b/core/tests/Drupal/Tests/Core/Entity/EntityUnitTest.php
@@ -98,6 +98,9 @@ protected function setUp() {
     $this->entityTypeId = $this->randomMachineName();
 
     $this->entityType = $this->getMock('\Drupal\Core\Entity\EntityTypeInterface');
+    $this->entityType->expects($this->any())
+      ->method('getListCacheTags')
+      ->willReturn(array($this->entityTypeId . '_list'));
 
     $this->entityManager = $this->getMock('\Drupal\Core\Entity\EntityManagerInterface');
     $this->entityManager->expects($this->any())
@@ -248,6 +251,7 @@ function setupTestLoad() {
       ->disableOriginalConstructor()
       ->setMethods($methods)
       ->getMock();
+
   }
 
   /**
@@ -392,13 +396,13 @@ public function testPostSave() {
     $this->cacheBackend->expects($this->at(0))
       ->method('invalidateTags')
       ->with(array(
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
     $this->cacheBackend->expects($this->at(1))
       ->method('invalidateTags')
       ->with(array(
         $this->entityTypeId . ':' . $this->values['id'], // Own cache tag.
-        $this->entityTypeId . 's', // List cache tag.
+        $this->entityTypeId . '_list', // List cache tag.
       ));
 
     // This method is internal, so check for errors on calling it only.
@@ -450,18 +454,14 @@ public function testPostDelete() {
       ->method('invalidateTags')
       ->with(array(
         $this->entityTypeId . ':' . $this->values['id'],
-        $this->entityTypeId . 's',
+        $this->entityTypeId . '_list',
       ));
     $storage = $this->getMock('\Drupal\Core\Entity\EntityStorageInterface');
+    $storage->expects($this->once())
+      ->method('getEntityType')
+      ->willReturn($this->entityType);
 
-    $entity = $this->getMockBuilder('\Drupal\Core\Entity\Entity')
-      ->setConstructorArgs(array($this->values, $this->entityTypeId))
-      ->setMethods(array('onSaveOrDelete'))
-      ->getMock();
-    $entity->expects($this->once())
-      ->method('onSaveOrDelete');
-
-    $entities = array($this->values['id'] => $entity);
+    $entities = array($this->values['id'] => $this->entity);
     $this->entity->postDelete($storage, $entities);
   }
 
diff --git a/core/tests/Drupal/Tests/Core/Entity/KeyValueStore/KeyValueEntityStorageTest.php b/core/tests/Drupal/Tests/Core/Entity/KeyValueStore/KeyValueEntityStorageTest.php
index 5c69511d371d33c931a9bcf020a0c4a42b958d49..adfefd59d1afa705b3e3c910486c211bb9a5f463 100644
--- a/core/tests/Drupal/Tests/Core/Entity/KeyValueStore/KeyValueEntityStorageTest.php
+++ b/core/tests/Drupal/Tests/Core/Entity/KeyValueStore/KeyValueEntityStorageTest.php
@@ -100,6 +100,9 @@ protected function setUpKeyValueEntityStorage($uuid_key = 'uuid') {
     $this->entityType->expects($this->atLeastOnce())
       ->method('id')
       ->will($this->returnValue('test_entity_type'));
+    $this->entityType->expects($this->any())
+      ->method('getListCacheTags')
+      ->willReturn(array('test_entity_type_list'));
 
     $this->entityManager = $this->getMock('Drupal\Core\Entity\EntityManagerInterface');
     $this->entityManager->expects($this->any())
@@ -426,7 +429,6 @@ public function testSaveContentEntity() {
     $this->keyValueStore->expects($this->never())
       ->method('delete');
     $entity = $this->getMockEntity('Drupal\Core\Entity\ContentEntityBase', array(), array(
-      'onSaveOrDelete',
       'toArray',
       'id',
     ));
diff --git a/core/modules/system/src/Tests/Entity/EntityCacheTagsTestBase.php b/core/modules/system/src/Tests/Entity/EntityCacheTagsTestBase.php
index b4e3fcedf8e8a7757334f1bcf6cbc10c9219a1de..3c981c12ab4b1c3a4db4431553564a39797a65bf 100644
--- a/core/modules/system/src/Tests/Entity/EntityCacheTagsTestBase.php
+++ b/core/modules/system/src/Tests/Entity/EntityCacheTagsTestBase.php
@@ -137,6 +137,19 @@ protected function getAdditionalCacheTagsForEntity(EntityInterface $entity) {
     return array();
   }
 
+  /**
+   * Returns the additional cache tags for the tested entity's listing by type.
+   *
+   * Necessary when there are unavoidable default entities of this type, e.g.
+   * the anonymous and administrator User entities always exist.
+   *
+   * @return array
+   *   An array of the additional cache tags.
+   */
+  protected function getAdditionalCacheTagsForEntityListing() {
+    return [];
+  }
+
   /**
    * Selects the preferred view mode for the given entity type.
    *
@@ -253,16 +266,19 @@ protected function createReferenceTestEntities($referenced_entity) {
    * Tests cache tags presence and invalidation of the entity when referenced.
    *
    * Tests the following cache tags:
    * - "<entity type>_view"
    * - "<entity type>:<entity ID>"
    * - "<referencing entity type>_view"
    * * - "<referencing entity type>:<referencing entity ID>"
    */
   public function testReferencedEntity() {
     $entity_type = $this->entity->getEntityTypeId();
     $referencing_entity_path = $this->referencing_entity->getSystemPath();
     $non_referencing_entity_path = $this->non_referencing_entity->getSystemPath();
     $listing_path = 'entity_test/list/' . $entity_type . '_reference/' . $entity_type . '/' . $this->entity->id();
+    $empty_entity_listing_path = 'entity_test/list_empty/' . $entity_type;
+    $nonempty_entity_listing_path = 'entity_test/list_labels_alphabetically/' . $entity_type;
 
     $render_cache_tags = array('rendered');
     $theme_cache_tags = array('theme:stark', 'theme_global_settings');
@@ -287,6 +303,21 @@ public function testReferencedEntity() {
       \Drupal::entityManager()->getViewBuilder('entity_test')->getCacheTag()
     );
 
+    // Generate the cache tags for all two possible entity listing paths.
+    // 1. list cache tag only (listing query has no match)
+    // 2. list cache tag plus entity cache tag (listing query has a match)
+    $empty_entity_listing_cache_tags = Cache::mergeTags(
+      $this->entity->getEntityType()->getListCacheTags(),
+      $theme_cache_tags,
+      $render_cache_tags
+    );
+    $nonempty_entity_listing_cache_tags = Cache::mergeTags(
+      $this->entity->getEntityType()->getListCacheTags(),
+      $this->entity->getCacheTag(),
+      $this->getAdditionalCacheTagsForEntityListing($this->entity),
+      $theme_cache_tags,
+      $render_cache_tags
+    );
 
     $this->pass("Test referencing entity.", 'Debug');
     $this->verifyPageCache($referencing_entity_path, 'MISS');
@@ -317,41 +348,62 @@ public function testReferencedEntity() {
     $this->verifyPageCache($listing_path, 'HIT', $tags);
 
 
+    $this->pass("Test empty listing.", 'Debug');
+    // Prime the page cache for the empty listing.
+    $this->verifyPageCache($empty_entity_listing_path, 'MISS');
+    // Verify a cache hit, but also the presence of the correct cache tags.
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT', $empty_entity_listing_cache_tags);
+
+
+    $this->pass("Test listing containing referenced entity.", 'Debug');
+    // Prime the page cache for the listing containing the referenced entity.
+    $this->verifyPageCache($nonempty_entity_listing_path, 'MISS');
+    // Verify a cache hit, but also the presence of the correct cache tags.
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT', $nonempty_entity_listing_cache_tags);
+
+
     // Verify that after modifying the referenced entity, there is a cache miss
-    // for both the referencing entity, and the listing of referencing entities,
-    // but not for the non-referencing entity.
+    // for every route except the one for the non-referencing entity.
     $this->pass("Test modification of referenced entity.", 'Debug');
     $this->entity->save();
     $this->verifyPageCache($referencing_entity_path, 'MISS');
     $this->verifyPageCache($listing_path, 'MISS');
+    $this->verifyPageCache($empty_entity_listing_path, 'MISS');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'MISS');
     $this->verifyPageCache($non_referencing_entity_path, 'HIT');
 
     // Verify cache hits.
     $this->verifyPageCache($referencing_entity_path, 'HIT');
     $this->verifyPageCache($listing_path, 'HIT');
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
 
 
     // Verify that after modifying the referencing entity, there is a cache miss
-    // for both the referencing entity, and the listing of referencing entities,
-    // but not for the non-referencing entity.
+    // for every route except the ones for the non-referencing entity and the
+    // empty entity listing.
     $this->pass("Test modification of referencing entity.", 'Debug');
     $this->referencing_entity->save();
     $this->verifyPageCache($referencing_entity_path, 'MISS');
     $this->verifyPageCache($listing_path, 'MISS');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
     $this->verifyPageCache($non_referencing_entity_path, 'HIT');
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT');
 
     // Verify cache hits.
     $this->verifyPageCache($referencing_entity_path, 'HIT');
     $this->verifyPageCache($listing_path, 'HIT');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
 
 
     // Verify that after modifying the non-referencing entity, there is a cache
-    // miss for only the non-referencing entity, not for the referencing entity,
-    // nor for the listing of referencing entities.
+    // miss only for the non-referencing entity route.
     $this->pass("Test modification of non-referencing entity.", 'Debug');
     $this->non_referencing_entity->save();
     $this->verifyPageCache($referencing_entity_path, 'HIT');
     $this->verifyPageCache($listing_path, 'HIT');
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
     $this->verifyPageCache($non_referencing_entity_path, 'MISS');
 
     // Verify cache hits.
@@ -361,7 +413,7 @@ public function testReferencedEntity() {
     if ($this->entity->getEntityType()->hasHandlerClass('view_builder')) {
       // Verify that after modifying the entity's display, there is a cache miss
       // for both the referencing entity, and the listing of referencing
-      // entities, but not for the non-referencing entity.
+      // entities, but not for any other routes.
       $referenced_entity_view_mode = $this->selectViewMode($this->entity->getEntityTypeId());
       $this->pass("Test modification of referenced entity's '$referenced_entity_view_mode' display.", 'Debug');
       $entity_display = entity_get_display($entity_type, $this->entity->bundle(), $referenced_entity_view_mode);
@@ -369,6 +421,8 @@ public function testReferencedEntity() {
       $this->verifyPageCache($referencing_entity_path, 'MISS');
       $this->verifyPageCache($listing_path, 'MISS');
       $this->verifyPageCache($non_referencing_entity_path, 'HIT');
+      $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+      $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
 
       // Verify cache hits.
       $this->verifyPageCache($referencing_entity_path, 'HIT');
@@ -380,17 +434,32 @@ public function testReferencedEntity() {
     if ($bundle_entity_type !== 'bundle') {
       // Verify that after modifying the corresponding bundle entity, there is a
       // cache miss for both the referencing entity, and the listing of
-      // referencing entities, but not for the non-referencing entity.
+      // referencing entities, but not for any other routes.
       $this->pass("Test modification of referenced entity's bundle entity.", 'Debug');
       $bundle_entity = entity_load($bundle_entity_type, $this->entity->bundle());
       $bundle_entity->save();
       $this->verifyPageCache($referencing_entity_path, 'MISS');
       $this->verifyPageCache($listing_path, 'MISS');
       $this->verifyPageCache($non_referencing_entity_path, 'HIT');
+      // Special case: entity types may choose to use their bundle entity type
+      // cache tags, to avoid having excessively granular invalidation.
+      $is_special_case = $bundle_entity->getCacheTag() == $this->entity->getCacheTag() && $bundle_entity->getEntityType()->getListCacheTags() == $this->entity->getEntityType()->getListCacheTags();
+      if ($is_special_case) {
+        $this->verifyPageCache($empty_entity_listing_path, 'MISS');
+        $this->verifyPageCache($nonempty_entity_listing_path, 'MISS');
+      }
+      else {
+        $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+        $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
+      }
 
       // Verify cache hits.
       $this->verifyPageCache($referencing_entity_path, 'HIT');
       $this->verifyPageCache($listing_path, 'HIT');
+      if ($is_special_case) {
+        $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+        $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
+      }
     }
 
 
@@ -403,6 +472,8 @@ public function testReferencedEntity() {
       $field_storage->save();
       $this->verifyPageCache($referencing_entity_path, 'MISS');
       $this->verifyPageCache($listing_path, 'MISS');
+      $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+      $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
       $this->verifyPageCache($non_referencing_entity_path, 'HIT');
 
       // Verify cache hits.
@@ -418,6 +489,8 @@ public function testReferencedEntity() {
       $field->save();
       $this->verifyPageCache($referencing_entity_path, 'MISS');
       $this->verifyPageCache($listing_path, 'MISS');
+      $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+      $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
       $this->verifyPageCache($non_referencing_entity_path, 'HIT');
 
       // Verify cache hits.
@@ -426,42 +499,63 @@ public function testReferencedEntity() {
     }
 
 
     // Verify that after invalidating the entity's cache tag directly,  there is
     // a cache miss for both the referencing entity, and the listing of
     // referencing entities, but not for the non-referencing entity.
     $this->pass("Test invalidation of referenced entity's cache tag.", 'Debug');
     Cache::invalidateTags($this->entity->getCacheTag());
     $this->verifyPageCache($referencing_entity_path, 'MISS');
     $this->verifyPageCache($listing_path, 'MISS');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'MISS');
     $this->verifyPageCache($non_referencing_entity_path, 'HIT');
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT');
 
     // Verify cache hits.
     $this->verifyPageCache($referencing_entity_path, 'HIT');
     $this->verifyPageCache($listing_path, 'HIT');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
+
+    // Verify that after invalidating the entity's list cache tag directly,
+    // there is a cache miss for both the empty entity listing and the non-empty
+    // entity listing routes, but not for other routes.
+    $this->pass("Test invalidation of referenced entity's list cache tag.", 'Debug');
+    Cache::invalidateTags($this->entity->getEntityType()->getListCacheTags());
+    $this->verifyPageCache($empty_entity_listing_path, 'MISS');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'MISS');
+    $this->verifyPageCache($referencing_entity_path, 'HIT');
+    $this->verifyPageCache($non_referencing_entity_path, 'HIT');
+    $this->verifyPageCache($listing_path, 'HIT');
+
+    // Verify cache hits.
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
 
 
     if (!empty($view_cache_tag)) {
       // Verify that after invalidating the generic entity type's view cache tag
       // directly, there is a cache miss for both the referencing entity, and the
-      // listing of referencing entities, but not for the non-referencing entity.
+      // listing of referencing entities, but not for other routes.
       $this->pass("Test invalidation of referenced entity's 'view' cache tag.", 'Debug');
       Cache::invalidateTags($view_cache_tag);
       $this->verifyPageCache($referencing_entity_path, 'MISS');
       $this->verifyPageCache($listing_path, 'MISS');
       $this->verifyPageCache($non_referencing_entity_path, 'HIT');
+      $this->verifyPageCache($empty_entity_listing_path, 'HIT');
+      $this->verifyPageCache($nonempty_entity_listing_path, 'HIT');
 
       // Verify cache hits.
       $this->verifyPageCache($referencing_entity_path, 'HIT');
       $this->verifyPageCache($listing_path, 'HIT');
     }
 
-    // Verify that after deleting the entity, there is a cache miss for both the
-    // referencing entity, and the listing of referencing entities, but not for
-    // the non-referencing entity.
+    // Verify that after deleting the entity, there is a cache miss for every
+    // route except for the the non-referencing entity one.
     $this->pass('Test deletion of referenced entity.', 'Debug');
     $this->entity->delete();
     $this->verifyPageCache($referencing_entity_path, 'MISS');
     $this->verifyPageCache($listing_path, 'MISS');
+    $this->verifyPageCache($empty_entity_listing_path, 'MISS');
+    $this->verifyPageCache($nonempty_entity_listing_path, 'MISS');
     $this->verifyPageCache($non_referencing_entity_path, 'HIT');
 
     // Verify cache hits.
@@ -473,6 +567,10 @@ public function testReferencedEntity() {
     $this->verifyPageCache($referencing_entity_path, 'HIT', $tags);
     $tags = Cache::mergeTags($render_cache_tags, $theme_cache_tags);
     $this->verifyPageCache($listing_path, 'HIT', $tags);
+    $tags = Cache::mergeTags($render_cache_tags, $theme_cache_tags, $this->entity->getEntityType()->getListCacheTags());
+    $this->verifyPageCache($empty_entity_listing_path, 'HIT', $tags);
+    $tags = Cache::mergeTags($tags, $this->getAdditionalCacheTagsForEntityListing());
+    $this->verifyPageCache($nonempty_entity_listing_path, 'HIT', $tags);
   }
 
   /**
@@ -253,16 +266,19 @@ protected function createReferenceTestEntities($referenced_entity) {
    *
    * Tests the following cache tags:
-   * - "<entity type>_view"
-   * - "<entity type>:<entity ID>"
-   * - "<referencing entity type>_view"
-   * * - "<referencing entity type>:<referencing entity ID>"
+   * - entity type view cache tag: "<entity type>_view"
+   * - entity cache tag: "<entity type>:<entity ID>"
+   * - entity type list cache tag: "<entity type>_list"
+   * - referencing entity type view cache tag: "<referencing entity type>_view"
+   * - referencing entity type cache tag: "<referencing entity type>:<referencing entity ID>"
    */
   public function testReferencedEntity() {
@@ -426,42 +499,63 @@ public function testReferencedEntity() {
 
 
-    // Verify that after invalidating the entity's cache tag directly,  there is
-    // a cache miss for both the referencing entity, and the listing of
-    // referencing entities, but not for the non-referencing entity.
+    // Verify that after invalidating the entity's cache tag directly, there is
+    // a cache miss for every route except the ones for the non-referencing
+    // entity and the empty entity listing.
     $this->pass("Test invalidation of referenced entity's cache tag.", 'Debug');
     Cache::invalidateTags($this->entity->getCacheTag());
diff --git a/core/modules/views/src/Plugin/views/cache/CachePluginBase.php b/core/modules/views/src/Plugin/views/cache/CachePluginBase.php
index a4626412529197a672a7725ca9021edb2c89a2ef..068197c7681dae30b7eac1c067722b08896b98ca 100644
--- a/core/modules/views/src/Plugin/views/cache/CachePluginBase.php
+++ b/core/modules/views/src/Plugin/views/cache/CachePluginBase.php
@@ -358,9 +358,9 @@ protected function getCacheTags() {
     $entity_information = $this->view->query->getEntityTableInfo();
 
     if (!empty($entity_information)) {
       // Add an ENTITY_TYPE_list tag for each entity type used by this view.
       foreach (array_keys($entity_information) as $entity_type) {
-        $tags[] = $entity_type . '_list';
+        $tags = Cache::mergeTags($tags, \Drupal::entityManager()->getDefinition($entity_type)->getListCacheTags());
       }
     }
 
@@ -358,9 +358,9 @@ protected function getCacheTags() {
 
     if (!empty($entity_information)) {
-      // Add an ENTITY_TYPE_list tag for each entity type used by this view.
+      // Add the list cache tags for each entity type used by this view.
       foreach (array_keys($entity_information) as $entity_type) {